/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* SCCS ID: @(#)knobconfig.c	1.3 5/12/94 */
/*
 * knobconfig.c,v
 * Revision 1.7  2002/08/02 15:39:30  jba
 * Added license information.
 *
 * Revision 1.6  2000/10/24 20:32:30  jba
 * Fixed comment lines.
 *
 * Revision 1.5  1996/02/16 22:19:03  saunders
 * Now honor the KnobDescription parameter. Printed out on each knob
 * assignment line.
 *
 * Revision 1.4  1995/10/10  21:40:33  jba
 * Portability changes.
 *
 * Revision 1.3  1994/06/17  17:42:58  saunders
 * Added code to handle drive limits. Provides indication of pegged knob now.
 *
 * Revision 1.2  1994/05/12  21:25:02  saunders
 * First fully function version in cvs addon environment.
 */
/*************************************************************************
 * FILE:      cknob.c
 * Author:    Claude Saunders
 * 
 * Purpose:   Program for configuring composite knobs
 *            
 *
 * Mods:      login mm/dd/yy: description
 *************************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <cadef.h>
#include <unistd.h>
#include <string.h>
#include <SDDS.h>          /* SDDS library header */
#include <libdev.h>        /* Device access library header */

static int debug = 0;      /* Internal debugging flag */

/* Function Prototypes */
int process_cmdline_args(int argc, char *argv[], char *spec);
void usage(char *name);

/* Data packet delivered by SunDials device */
typedef struct inputevent {
  short device;              /* which knob (0-7) */
  short unused;
  int delta;                 /* change in position since last scan */
  struct timeval time;       /* system time at which scan took place */
} Event;

main(int argc, char *argv[]) {
  /* SDDS related variables */
  SDDS_TABLE table;
  int tablenum;
  char **name;                /* device name or process variable */
  char **type;                /* dev or pv */
  char **message;             /* message for device control */
  char **knobDescription;     /* description of knob assignment */
  char **componentname[8];    /* composite device constituents */
  double *weight;             /* relative weight */
  double *gain;               /* knob gain (default of 10.0) */

  char specfilename[255];     /* specification file name */
  char *text, *contents;      /* text and contents fields of SDDS descrip. */
  long nrows;
  long i, j, n;
  char buf[255];

  /* SunDials/Knob related variables */
  int dbfd;          /* SunDials device file descriptor */
  int nread;         /* bytes read from SunDials device */
  Event event;       /* format of 16 byte SunDials packet */
  int delta;      /* SunDials knob delta value */
  double norm[8];    /* normalized knob delta value */
  double current[8];
  int offset[8];
  int tick;
  int more = 1;
  double prec;
  double wdelta;
  int knob = 0;

  /* Device access related variables */
  int celm[8];
  devnamearr mynames;
  int status;
  chid *devchid[8][MAX_COMPONENTS];  /* one array of chids for each dial */
  double val[8][MAX_COMPONENTS];     /* for composite device set and read */
  double wvec[8][MAX_COMPONENTS];
  double gainvec[8];
  int pegged[8];
  struct dbr_ctrl_double ctrl_struct[8][MAX_COMPONENTS];

  process_cmdline_args(argc, argv, specfilename);

  fprintf(stdout,"*******************************\n");
  fprintf(stdout,"** Composite Knob Assignment **\n");
  fprintf(stdout,"*******************************\n");

#ifdef SUNOS4
  if ((dbfd = open("/dev/dialbox",0400)) == -1) {
    perror("Open failed for /dev/dialbox:");
    exit(1);
  }
  status = flock(dbfd, (LOCK_EX | LOCK_NB));
#endif
#ifdef SOLARIS
  if ((dbfd = open("/dev/bd",0400)) == -1) {
    perror("Open failed for /dev/bd:");
    exit(1);
  }
  status = lockf(dbfd, F_TLOCK, 0);
#endif
  if (status == -1) {
    fprintf(stderr,"%s: unable to access dialbox, probably in use\n",argv[0]);
    exit(1);
  }

  if (!SDDS_InitializeInput(&table, specfilename)) {
    fprintf(stderr,"%s: SDDS error: unable to initialize with file %s\n",
	    argv[0], specfilename);
    exit(1);
  }
  
  while (((tablenum = SDDS_ReadTable(&table)) != -1) && (knob < 8)) {
    if (tablenum == 0) {
      fprintf(stderr,"%s: SDDS error: unable to read data table from %s\n",
	      argv[0], specfilename);
      exit(1);
    }
    SDDS_SetColumnFlags(&table, 1);
    SDDS_SetRowFlags(&table, 1);
    
    contents = NULL;
    if (!SDDS_GetDescription(&table, NULL, &contents)) {
      fprintf(stderr,"%s: SDDS error: unable to get description from file %s\n"
	      ,argv[0], specfilename);
      exit(1);
    }
    if (contents==NULL || strcmp(contents,"knob assignment specification")) {
      fprintf(stderr,"%s: SDDS error: file %s is not a knob assignment specification\n",argv[0],specfilename);
      exit(1);
    }
    
    if ((nrows = SDDS_CountRowsOfInterest(&table)) == -1) {
      fprintf(stderr,"%s: SDDS error: unable to get row count\n",argv[0]);
      exit(1);
    }
    
    if ((name = SDDS_GetParameter(&table, "ControlName", NULL)) == NULL) {
      fprintf(stderr,"%s: SDDS error: unable to read ControlName parameter\n",
	      argv[0]);
      exit(1);
    }
    
    if ((type = SDDS_GetParameter(&table, "ControlType", NULL)) == NULL) {
      fprintf(stderr,"%s: SDDS error: unable to read ControlType parameter\n",
	      argv[0]);
      exit(1);
    }

    if ((message = SDDS_GetParameter(&table, "ControlMsg", NULL)) == NULL) {
      /* default of set */
      message = (char **)malloc(sizeof(char **));
      *message = malloc(sizeof(char)*255);
      strcpy(*message,"set");
    }

    if ((gain = SDDS_GetParameter(&table, "Gain", NULL)) == NULL) {
      fprintf(stderr,"%s: SDDS error: unable to read Gain parameter\n",
	      argv[0]);
      exit(1);
    }
    gainvec[knob] = *gain;

    if ((knobDescription = SDDS_GetParameter(&table, "KnobDescription", NULL))
	== NULL) {
      knobDescription = (char **)malloc(sizeof(char **));
      *knobDescription = malloc(sizeof(char)*255);
      strcpy(*knobDescription,"no description");
    }

     /* Open connections to process variables and intialize data */
    if (!(strcmp(*type,"dev"))) {
      status = devCount(*name,&celm[knob],&mynames);
      if (devError(status)) {
	fprintf(stderr,"%s: error: device %s does not exist\n",argv[0],*name);
	exit(1);
      }
      if (celm[knob] == 1) {
	wvec[knob][0] = 1.0;
	componentname[knob] = (char **)malloc(sizeof(char *));
	*componentname[knob] = malloc(255);
	strcpy(*componentname[knob], *name);
      } else {
	if (nrows != celm[knob]) {
	  fprintf(stderr,
		  "%s: error: you must supply %d weights for device %s\n",
		  argv[0],celm[knob],*name);
	  exit(1);
	}
	if ((componentname[knob]= SDDS_GetColumn(&table,"ControlName"))==NULL){
	  fprintf(stderr,"%s: SDDS error: unable to read ControlName column\n",
		  argv[0]);
	  exit(1);
	}
	if ((weight = SDDS_GetColumn(&table, "Weight"))==NULL) {
	  fprintf(stderr,"%s: SDDS error: unable to read Weight column\n",
		  argv[0]);
	  exit(1);
	}
	/* copy weight vector */
	for (i=0 ; i < celm[knob] ; i++)
	  wvec[knob][i] = weight[i];
      }
      /* make connection */
      for (i=0 ; i < celm[knob] ; i++) {
	status = devChid(componentname[knob][i],*message,&(devchid[knob][i]));
	if (devError(status)) {
	  fprintf(stderr,"%s: error: unable to connect with %s\n",
		  argv[0],componentname[knob][i]);
	  exit(1);
	}
	/* collect control limit information on this channel */
	status = ca_get(DBR_CTRL_DOUBLE, *(devchid[knob][i]), 
			&(ctrl_struct[knob][i]));
	if (status != ECA_NORMAL) {
	  fprintf(stderr,"%s: unable to get control limits on %s\n",argv[0],
		  componentname[knob][i]);
	  exit(1);
	}
	status = ca_pend_io(5.0);
	if (status != ECA_NORMAL) {
	  fprintf(stderr,"%s: unable to get control limits on %s\n",argv[0],
		  componentname[knob][i]);
	  exit(1);
	}
      }
      fprintf(stdout,"Dev %24s --> knob %d :: %34s\n",*name,knob+1,*knobDescription);
    } else {
      celm[knob] = 1;
      wvec[knob][0] = 1.0;
      componentname[knob] = (char **)malloc(sizeof(char *));
      *componentname[knob] = malloc(255);
      strcpy(*componentname[knob], *name);

      devchid[knob][0] = (chid *)malloc(sizeof(chid));

      status = ca_search(*name,devchid[knob][0]);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to connect with %s\n",argv[0],*name);
	exit(1);
      }
      status = ca_pend_io(5.0);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to connect with %s\n",argv[0],*name);
	exit(1);
      }
      /* collect control limit information on this channel */
      status = ca_get(DBR_CTRL_DOUBLE, *(devchid[knob][0]), 
		      &(ctrl_struct[knob][0]));
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to get control limits on %s\n",argv[0],
		componentname[knob][0]);
	exit(1);
      }
      status = ca_pend_io(5.0);
      if (status != ECA_NORMAL) {
	fprintf(stderr,"%s: unable to get control limits on %s\n",argv[0],
		componentname[knob][0]);
	exit(1);
      }
      fprintf(stdout,"Pv  %24s --> knob %d :: %34s\n",*name,knob+1,*knobDescription);
    }

    knob++;
  }

  /* intialize vars */
  for (i=0 ; i < 8 ; i++) {
    current[i] = 0.0;
    offset[i] = 0;
    pegged[i] = 0;
  }
  tick = 0;
  /* Prec determines how many "puts" a given amount 
     of knob rotation will produce. A higher value
     results in more puts.
  */
  prec = 15.0;

  fprintf(stdout,"OFFSETS:\n");
  for (j=0 ; j < knob ; j++) 
    fprintf(stdout,"k%d: %d%s ",j+1,offset[j],(pegged[j]?"P":" "));
  fflush(stdout);

  /* handle knob box output */
  while (more) {
    nread = read(dbfd,(char *)&event,16);   /* read one knob data packet */
    if (nread == 16) {                      /* knob data received */
      n = event.device & 0xff;              /* get knob number */

      if (n <= knob) {
	delta = event.delta;
	norm[n] = delta/65535.0;      /* normalize delta */
	norm[n] = norm[n] * prec;     /* scale delta     */
	current[n] = current[n] + norm[n];
	if (current[n] > 1.0) {
	  current[n] = current[n] - 1.0;
	  wdelta = 1.0;
	  tick = 1;
	  offset[n]++;
	}
	else if (current[n] < -1.0) {
	  current[n] = current[n] + 1.0;
	  wdelta = -1.0;
	  tick = 1;
	  offset[n]--;
	}
	if (tick) {
	  for (i=0 ; i < celm[n] ; i++) {
	    status = ca_get(DBR_DOUBLE,*(devchid[n][i]),&(val[n][i]));
	    if (status != ECA_NORMAL) {
	      fprintf(stderr,"%s: error: unable to read %s\n",
		      argv[0],componentname[n][i]);
	      exit(1);
	    }
	  }
	  status = ca_pend_io(5.0);
	  if (status != ECA_NORMAL) {
	    fprintf(stderr,"%s: error: unable to read %s\n",
		    argv[0],componentname[n][i]);
	    exit(1);
	  }

	  for (i=0 ; i < celm[n] ; i++) {
	    val[n][i] = val[n][i] + ((wdelta * gainvec[n]) * wvec[n][i]);

	    /* Check control limits. Mark knob as pegged if any exceeded. */
	    if (!pegged[n] && 
		(val[n][i] > ctrl_struct[n][i].upper_ctrl_limit ||
		 val[n][i] < ctrl_struct[n][i].lower_ctrl_limit))
	      pegged[n] = 1;
	    else if (pegged[n] && 
		     val[n][i] <= ctrl_struct[n][i].upper_ctrl_limit &&
		     val[n][i] >= ctrl_struct[n][i].lower_ctrl_limit)
	      pegged[n] = 0;
	  }

	  /* Correct offset if knob was marked as pegged. */
	  if (pegged[n]) {
	    if (wdelta < 0.0)
	      offset[n]++;
	    else
	      offset[n]--;
	  }
	  else {
	    for (i=0 ; i < celm[n] ; i++) {
	      status = ca_put(DBR_DOUBLE,*(devchid[n][i]),&(val[n][i]));
	      if (status != ECA_NORMAL) {
		fprintf(stderr,"%s: error: unable to set %s\n",
			argv[0],componentname[n][i]);
		exit(1);
	      }
	    }
	    status = ca_pend_io(5.0);
	    if (status != ECA_NORMAL) {
	      fprintf(stderr,"%s: error: unable to set %s\n",
		      argv[0],componentname[n][i]);
	      exit(1);
	    }
	  }
	  fprintf(stdout,"\015");
	  for (j=0 ; j < knob ; j++) 
	    fprintf(stdout,"k%d: %d%s ",j+1,offset[j],
		    (pegged[j]?"P":" "));
	  fflush(stdout);
	  tick = 0;
	}
      }
    } else if (nread == 0) {                /* EOF reached (not likely) */
      more = 0;
    } else if (nread == -1) {               /* Error */
      perror("error reading sundials:");
    } else if (nread < 16) {
      fprintf(stderr,"cknob: not reading whole dial packets\n");
    }
    fflush(stdout); 
  }
  exit(0);
  
}

/*************************************************************************
 * FUNCTION : 
 * PURPOSE  : 
 *            
 * ARGS in  : 
 *            
 * ARGS out : none
 * GLOBAL   : nothing
 * RETURNS  : 0 - time values valid 
 *            -1 - at least one time value is invalid
 ************************************************************************/


/*************************************************************************
 * FUNCTION : process_cmdline_args()
 * PURPOSE  : Get command line arguments and set flags appropriately.
 * ARGS in  : argc, argv - standard unix
 * ARGS out : spec, query, configure, stop
 * GLOBAL   : nothing
 * RETURNS  : 0
 ************************************************************************/  
int
process_cmdline_args(int argc, char *argv[], char *spec)
{
  int i;
  int cmderror = 0;
  void usage();

  *spec = '\0';

  /* Process command line options */
  if (argc < 2 || argc > 2)
    cmderror++;
  else {
    for (i=1 ; i < argc ; i++) {
	strcpy(spec,argv[i]);
    }
  }

  if (cmderror) {
    usage(argv[0]);
    exit(1);
  }
  return(0);
}

/*************************************************************************
 * FUNCTION : usage()
 * PURPOSE  : Describes command line options available to client.
 * ARGS in  : char *name - name of executable program
 * ARGS out : none
 * GLOBAL   : nothing
 * RETURNS  : none
 ************************************************************************/  
void usage(char *name)
{
  fprintf(stderr,"usage: %s spec-file\n",name);
  fprintf(stderr,"purpose:\tConfigures knobs.\n");
  fprintf(stderr,"\t\tAssigns knobs to devices and process variables\n");
  fprintf(stderr,"\t\taccording to spec-file. Assignments are active\n");
  fprintf(stderr,"\t\tonly during program execution. In general, it \n");
  fprintf(stderr,"\t\tis best not to run this utility in background.\n");
  fprintf(stderr,"Sample spec-file. Note that ControlMessage parm is optional:\n");
  fprintf(stderr,"SDDS1 \n");
  fprintf(stderr,"&description \n");
  fprintf(stderr,"        text=\"Knob Assignment Specification\" \n");
  fprintf(stderr,"        contents=\"knob assignment specification\" \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&parameter name=ControlName \n");
  fprintf(stderr,"           type=string  \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&parameter name=ControlType \n");
  fprintf(stderr,"           type=string \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&parameter name=ControlMsg \n");
  fprintf(stderr,"           type=string \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&parameter name=ControlUnits \n");
  fprintf(stderr,"           type=string \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&parameter name=Gain \n");
  fprintf(stderr,"           type=double \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&column name=ControlName \n");
  fprintf(stderr,"        type=string \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&column name=Weight \n");
  fprintf(stderr,"        type=double \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"&data   mode=ascii \n");
  fprintf(stderr,"        no_row_counts=1 \n");
  fprintf(stderr,"&end \n");
  fprintf(stderr,"COMPDEVICENAME \n");
  fprintf(stderr,"dev \n");
  fprintf(stderr,"set \n");
  fprintf(stderr,"tune \n");
  fprintf(stderr,"0.01 \n");
  fprintf(stderr,"P:Q1  1.0000 \n");
  fprintf(stderr,"P:Q2  1.1134 \n");
  fprintf(stderr," \n");
  fprintf(stderr,"PTB:Q3 \n");
  fprintf(stderr,"dev \n");
  fprintf(stderr,"set \n");
  fprintf(stderr,"Amps \n");
  fprintf(stderr,"0.1 \n");
  fprintf(stderr," \n");
  fprintf(stderr,"PTB:Q4:CurrentAO \n");
  fprintf(stderr,"pv \n");
  fprintf(stderr,"- \n");
  fprintf(stderr,"Amps \n");
  fprintf(stderr,"0.1 \n");
  fprintf(stderr," \n");

}
